(*#line 117 "wordgraph.nw"*)signature WORDGRAPH =
    sig
        type token = char
        type state = int

        val start  : state
        val reject : state

        type dawg = {sink : state,
                     states : state list,
                     max : int, 
                     pos : int, 
                     tokens : token list,
                     positions : state -> int list, 
                     suffix : state -> state, 
                     trans : state * (token * bool) -> state}

        val mkDawg : string -> dawg
        val updateDawg : token * dawg -> dawg
    end

(*#line 157 "wordgraph.nw"*)structure Wordgraph =
   struct 
      type token = char
      type state = int

      val start = 0
      val (SOME (reject : state)) = Int.maxInt

      type dawg = {sink : state,
                   states : state list,
                   max : int,
                   pos : int,
                   tokens : token list, 
                   positions : state -> int list,
                   suffix : state -> state,
                   trans : state * (token * bool) -> state}

(*#line 179 "wordgraph.nw"*)      val initG : dawg = 
         {sink = 0,
          states = [0],
          max = 0,
          pos = 0,
          positions = (fn q => if q=0 then [0] else []),
          tokens = [],
          suffix = (fn q => reject),
          trans = (fn (q, (a:token,b:bool)) => reject)}

(*#line 214 "wordgraph.nw"*)      fun updateF ((arg,value),f) = 
         (fn x => if x=arg then value else (f x))

      fun add a [] = [a] 
        | add a (b::bs) = if a = b then b::bs else b::(add a bs)

(*#line 225 "wordgraph.nw"*)      fun updateDawg (a,G : dawg) : dawg =  
         let
(*#line 230 "wordgraph.nw"*)            val newSink = (#max G) + 1    
            val newPos = (#pos G) + 1     
            val newTokens = add a (#tokens G)

(*#line 240 "wordgraph.nw"*)            fun suffixChain q = 
               if not (hasTransition (q,a)) andalso not (q = reject)
                  then q::(suffixChain (#suffix G q))
               else []

            and hasTransition (q,a) =
               hasPrimaryTransition (q,a) orelse hasSecondaryTransition (q,a)
            and hasPrimaryTransition (q,a) = 
               not (#trans G (q,(a,true)) = reject)
            and hasSecondaryTransition (q,a) =
               not (#trans G (q,(a,false)) = reject)
               
            val chain = suffixChain (#sink G) 

(*#line 259 "wordgraph.nw"*)            fun toNewSink p = if p = (#sink G) then ((p,(a,true)),newSink)
                              else ((p,(a,false)),newSink)
            val newTrans = 
               foldl updateF (#trans G) (map toNewSink chain)

(*#line 273 "wordgraph.nw"*)         in 
            if (List.last chain) = start then

(*#line 284 "wordgraph.nw"*)               {sink = newSink,
                states = add newSink (#states G), 
                max = newSink,
                pos = newPos,
                tokens = newTokens,
                positions = (fn q => if q = start
                                        then newPos :: (#positions G q)
                                     else if q = newSink then [newPos] 
                                          else #positions G q),
                suffix = (fn q => if q = newSink then start else (#suffix G q)),
                trans = newTrans} 

(*#line 301 "wordgraph.nw"*)            else
               let
                  val p = #suffix G (List.last chain)
               in 
(*#line 310 "wordgraph.nw"*)                  if hasPrimaryTransition (p,a) then 
                              (*                      delta([r],a) = [sa] *)
                     let
                        val pa = #trans G (p,(a,true))
                        fun newSuff q = if q = newSink then pa else (#suffix G q)
                        
(*#line 324 "wordgraph.nw"*)                        fun newPositions q = 
                           let 
                              fun rootPath q acc = 
                                 if q = start then (q::acc) 
                                 else rootPath (newSuff q) (q::acc)

                              fun pos' s = if s = newSink then []
                                           else (#positions G s)
                              fun upd (q,f) = updateF ((q, newPos::(pos' q)),f) 
                              val g = (foldr upd pos' (rootPath newSink []))
                           in
                              g q
                           end
                     in 
                        {sink = newSink,
                         states = add newSink (#states G), 
                         max = newSink,
                         pos = newPos,
                         tokens = newTokens,
                         positions = newPositions,
                         suffix = newSuff,
                         trans = newTrans}
                     end

(*#line 362 "wordgraph.nw"*)                  else 
                     let 
                        val pa = #trans G (p,(a,false)) (* [max[pa]_s]_sa *)
                        val newState = newSink + 1      (* [pa]_sa *)

                        fun suffixChain' q = 
                           if (hasTransition (q,a)) 
                              andalso not (q = reject)
                              then q::(suffixChain' (#suffix G q))
                           else []
                        val chain' = suffixChain' (#suffix G p) (* ! *)

                        val newTrans = 
                           let 

(*#line 380 "wordgraph.nw"*)                              fun outgoing (b,f) = 
                                 if hasPrimaryTransition (pa,b) 
                                    then updateF (((newState,(b,false)),
                                               #trans G (pa,(b,true))),f) 
                                 else if hasSecondaryTransition (pa,b) 
                                         then updateF (((newState,(b,false)),
                                                        #trans G (pa,(b,false))),f)
                                      else f
                              val outTr1 = foldr outgoing (#trans G) (#tokens G)

(*#line 394 "wordgraph.nw"*)                              val outTr2 = updateF (((#sink G,(a,true)),newSink),outTr1)

                              fun outgoingNew (b,f) = 
                                 if pa = #sink G andalso b = a 
                                    then updateF (((newState,(b,false)), newSink), f)
                                 else f

                              val outTr = foldr outgoingNew outTr2 (#tokens G)

(*#line 406 "wordgraph.nw"*)                              fun incoming (q,f) =
                                 if f (q,(a,false)) = pa then 
                                    updateF (((q,(a,false)),newState),f)
                                 else f
                              val inTr1 = foldl incoming outTr chain'

(*#line 415 "wordgraph.nw"*)                              val inTr2 = updateF (((p,(a,false)),reject),inTr1)
                              val inTr = updateF (((p,(a,true)),newState),inTr2)

(*#line 422 "wordgraph.nw"*)                           in
                              (foldl updateF inTr (map toNewSink chain))
                           end

(*#line 428 "wordgraph.nw"*)                        fun newSuff q = 
                           if q = newState then #suffix G pa
                           else if q = pa then newState 
                                else if q = newSink then newState 
                                     else #suffix G q             

(*#line 439 "wordgraph.nw"*)                        fun newPositions q = 
                           let 
                              fun rootPath q acc = 
                                 if q = start then (q::acc) 
                                 else rootPath (newSuff q) (q::acc)

                              fun pos' s = if s = newSink then [newPos]
                                           else if s = newState then 
                                               newPos::(#positions G pa)
                                                else (#positions G s)

                              fun upd (q,f) = if q <= #max G then 
                                 updateF ((q, newPos::(pos' q)),f) else f
                              val g = (foldr upd pos' (rootPath newSink [])) 
                           in
                              g q
                           end

(*#line 461 "wordgraph.nw"*)                        fun replace (n,ms) (q::qs) acc = 
                           if q = n then (rev qs) @ (ms @ (rev acc))
                           else replace (n,ms) qs (q::acc)
                        val newStates =
                           replace (pa,[newState,pa]) (newSink :: (rev (#states G))) []
                     in
                        {sink = newSink,
                         states = newStates, 
                         max = #max G + 2,
                         pos = newPos,
                         tokens = newTokens,
                         positions = newPositions,
                         suffix = newSuff,
                         trans = newTrans}
                     end
               end
         end
      fun mkDawg text = foldl updateDawg initG (explode text)

      fun state (G:dawg) w = 
          let
              fun hasPrimaryTransition (q,a) = 
                  not (#trans G (q,(a,true)) = reject)
              and hasSecondaryTransition (q,a) =
                  not (#trans G (q,(a,false)) = reject)
              fun delta (a,q) = 
                  if hasPrimaryTransition (q,a) then (#trans G (q,(a,true)))
                  else (#trans G (q,(a,false)))
          in
              foldl delta start (explode w)
          end
      and occurs G w = #positions G (state G w)
 end

